--[[
************************************************************************
Project				: SKG
Author				: zhinjio
Project Revision	: r122
Project Date		: 20081223075030

File				: SKG_Comm.lua
Commit Author		: zhinjio
Commit Revision		: 120
Commit Date			: 20081222091749
************************************************************************
Description	:
	This file handles all the listeners for incoming communications, and
	routes the commands to the appopriate functions.

	It should be noted that the addon does not do really *anything* just
	native to itself. It almost always calls a COMM so that it essentially
	will always be a client to its own COMM calls.

TODO		:

************************************************************************
(see bottom of file for changelog)
************************************************************************
--]]
local SKG = LibStub( "AceAddon-3.0"):GetAddon( "SKG" )
local SKGL = LibStub( "AceLocale-3.0"):GetLocale( "SKG" )

-- My messages are all going to be complex types, and very rarely just strings. As
-- such, I need to serialize all my sends and deserialize all receives. In addition,
-- my receiving function is going to always treat the first argument as a "token",
-- indicating the type of communication, and what function to send the payload to

local SKG_OCR = {}

function SKG_SendComm( token, payload, distribution, target )
	local newPayload = SKG:Serialize( token, payload )
	SKG:SendCommMessage( SKG_CHANNEL_PREFIX, newPayload, distribution, target, "BULK" )
end

function SKG:OnCommReceived( prefix, message, distribution, sender )
	SKG_DoIt( sender, SKG:Deserialize( message ) )
end

function SKG_DoIt( sender, result, ... )
	if ( result ) then
		local token, payload = ...
		SKG_Debug( "OCR : " .. token )
		SKG_OCR[token]( sender, payload )
	else
		local errormsg = ...
		SKG_Debug( "Unable to Deserialize message : " .. errormsg ) 
	end
end

SKG_OCR.DEBUG = function( sender, message )
	SKG_Debug( "Comm: Debug (" .. sender .. ") : " .. message )
end

------------------------------------------------
-- Syncronization Commands 
------------------------------------------------

SKG_OCR.SYNSS = function( sender, message )
	-- SYNSS
	-- Payload: { TID, ListName, ListHash }
	-- Description: Snapshot the current lists and wipe our Log

	-- Details:
	-- SYNSS events are sent to "catalog" a save point for lists once the syncers
	-- agree they all have a correct list. It can also be done to reduce the space
	-- of the log is taking up.
	--
	-- General logic: if your hash for the named list is the same as the one specified
	-- in the payload, you save a copy of the named list to the Snapshot structure
	-- and initialize the Log for that list to the snapshot event (clearing it out)
	--
	-- If you do NOT have the same hash for the named list, flag an error to the user
	-- indicating that they will probably need to do a forced sync from someone to
	-- get the correct list. (it probably means they don't have the right list).

	-- If the receiving client is read-only, this can be safely ignored.

	SKG_Debug( "List Snapshot received" )
	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- We're not read-only, do that thang.
		local TID, ListName, ListHash = unpack( message, 1, 3 )
		local db = SKG.db.profile

		-- Check the hash. If our hash tastes like their hash, commit
		-- need to check the has using latestTID, not the TID sent
		local latestTID = SKG_GetLatestTID( db.Log[ListName] )
		local LH, SH = SKG_HashRollList( latestTID, ListName )
		if ( LH == ListHash ) then
			-- do eeet
			db.Snapshot[ListName] = {}
			db.Snapshot[ListName] = SKG_tblCopy ( db.RollLists[ListName] )
			db.Log[ListName] = {}
			db.SnapshotLM[ListName] = {}
			db.SnapshotLM[ListName] = SKG_tblCopy ( db.LastMod[ListName] )
			SKG_AddLogEntry( TID, ListName, "SYNSS", ListHash, ListHash )
			SKG_Print( SKGL["SKG_MSG_SNAPSHOTCOMPLETE"] )
		else
			-- Popup some warning that the lists are not synced, and they should
			-- get a sync from someone.
			StaticPopup_Show( "SKG_SYNSS_WITH_BAD_HASH" )
		end
	end
end

SKG_OCR.SYNRK = function( sender, message )
	-- SYNRK
	-- Payload: List
	-- Description: Request Sync Key for "List" (reply to sender)

	SKG_Debug( "Comm: Sync Key Request received" )
	local db = SKG.db.profile

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- TODO: Security: Put in a check to see if we're a syncer. If not, don't do anything
		if ( message == "Players" ) then
			local longHash, shortHash = SKG_HashRollList( 0, "Players" )
			SKG_SendComm( "SYNSK", longHash .. shortHash, "WHISPER", sender )
		else
			local latestTID = 0
			local longHash, shortHash = "", ""
			-- Make sure we have that list!!
			if ( db.RollLists[message] ) then
				latestTID = SKG_GetLatestTID( db.Log[message] )
				longHash, shortHash = SKG_HashRollList( latestTID, message )
			else
				shortHash = "nil"
			end
			SKG_SendComm( "SYNSK", shortHash, "WHISPER", sender )
		end
	end
end

SKG_OCR.SYNSK = function( sender, message )
	-- SYNSK
	-- Payload: ListHash
	-- Description: Send Sync Key for "List"

	SKG_Debug( "Comm: Sync Key Send received" )

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- Don't do anything if I'm the sender of the request
		if not ( sender == SKG.me ) then
			SKG_currentSyncItems = SKG_currentSyncItems + 1
			local currButton = SKG_currentSyncItems
	
			SKG_SYF_NameListItem[currButton]:SetText( sender )
			SKG_SYF_NameListItem[currButton]:Show()
			SKG_SYF_NameListItem[currButton]:Enable()
			SKG_SYF_HashListItem[currButton]:SetText( message )
			SKG_SYF_HashListItem[currButton]:Show()
		end
	end
end

SKG_OCR.SYNRL = function( sender, message )
	-- SYNRL
	-- Payload: { List, ListHash }
	-- Description: Request whispered from "sender" to send them your copy of
	--		"List". This request requires no acknowledgement. Simply send them
	--		the requested list if you have it. The ListHash is sent to compare
	--		to our own. If the ListHash matches, there's no work to be done.

	SKG_Debug( "Comm: Sync List Request" )

	-- I can't send myself this request, so don't bother checking...
	-- don't bother recording requests, since they don't change any data
	-- just send over the list

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		local listName, listHash = unpack( message, 1, 2 )
		local db = SKG.db.profile

		if ( listName == "Players" ) then
			SKG_Debug( "SYNRL: list is Players" )
			if ( next( db.Players ) == nil ) then
				SKG_Debug( "SYNRL: seems Players is empty??" )
				-- I don't think I want to send nil. Just don't do anything for now
				-- SKG:SendCommMessage( "WHISPER", sender, "SYNSL", nil )
			else
				-- don't bother comparing listhashes for Players. Just send it
				SKG_Debug( "SYNRL: Players is not empty" )
				local newTID = SKG_CreateTID( 0, 0, 0 )
				SKG_Print( string.format( SKGL["SKG_MSG_SENDINGLIST"], "Players", sender ) )
				SKG_SendComm( "SYNSL", { newTID, "Players", db.Players, db.Alts }, "WHISPER", sender )
				SKG_Print( string.format( SKGL["SKG_MSG_FINISHEDSEND"], "Players", sender ) )
			end
		else
			-- otherwise, it should be a rolllist
			-- do we have that rolllist?
			SKG_Debug( "SYNRL: list is not Players" )
			if ( db.RollLists[listName] ) then
				-- ok, we have the same list. Does our hash match his hash?
				SKG_Debug( "SYNRL: we have that list" )
				local maxTID = SKG_GetLatestTID( db.Log[listName] )
				local ourHash = SKG_HashRollList( maxTID, listName )
				if ( ourHash == listHash ) then
					SKG_Debug( "SYNRL: hashes same" )
				else
					-- Send along the list and its log
					SKG_Debug( "SYNRL: hashes not same" )
					local playerHash = SKG_HashRollList( 0, "Players" )
					SKG_Debug( "SYNRL: created playerHash" )
					local newTID = SKG_CreateTID( 0, 0, 0 )
					SKG_Debug( "SYNRL: created newTID" )
					SKG_Print( string.format( SKGL["SKG_MSG_SENDINGLIST"], listName, sender ) )
					SKG_SendComm( "SYNSL", { newTID, listName, db.RollLists[listName],
						db.Log[listName], db.LastMod[listName], db.Snapshot[listName],
						db.SnapshotLM[listName], playerHash }, "WHISPER", sender )
					SKG_Debug( "SYNRL: sent message SYNSL" )
					SKG_Print( string.format( SKGL["SKG_MSG_FINISHEDSEND"], listName, sender ) )
				end
			end
		end
	end
end

SKG_OCR.SYNSL = function( sender, message )
	-- SYNSL
	-- Payload: { TID, listName, Table1, Table2, [ PlayerHash ] }
	-- Description: This can only be received by someone if they had previously
	--		requested it, so there's no need to send any authorization requests.
	--		This will be returned in one of two ways:
	--			{ TID, "Players", PlayersTable, AltsTable }
	--			{ TID, listName, listTable, listLogTable, 
	--				LMTable, snapTable, snapLMTable, playerHash }

	SKG_Debug( "Comm: Received Sent List (answer from SYNRL)" )

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		local db = SKG.db.profile
		local TID, listName = unpack( message, 1, 2 )
		if ( listName == "Players" ) then
			SKG_Debug( "SYNSL: Players list" )
			local playerList, altList = unpack( message, 3, 4 )

			local newPL = db.Players
			for hash, plInfo in pairs( playerList ) do
				if ( not newPL[hash] ) then
					-- we got a new one, add it in.
					newPL[hash] = plInfo
					SKG_Debug( "Adding " .. hash .. " (" .. newPL[hash].name .. "/" .. newPL[hash].class .. ")" )
				end
			end
			local lHash, sHash = SKG_HashRollList( 0, "Players" )

			local newAL = db.Alts
			for hash, plAlts in pairs( altList ) do
				if ( not newAL[hash] ) then
					-- I have a player in the alts list that isn't in mine. Make sure it exists
					newAL[hash] = {}
				end
				for altIndex, altInfo in pairs( plAlts ) do
					-- since the lists could be completely different for the two people,
					-- make sure you look at the whole list to see if the alt already exists.
				
					-- so... is this entry in my list already?
					local notfound = true
					local i = 1
					while ( ( i <= #newAL[hash] ) and ( notfound == true ) ) do
						if ( newAL[hash][i].name == plAlts[altIndex].name ) then
							notfound = false
						end
						i = i + 1
					end
					if ( notfound == true ) then
						tinsert( newAL[hash], plAlts[altIndex] )
					end
				end
			end

			SKG_AddLogEntry( TID, "SKG", "SYNSL", 0, lHash .. sHash )
		else
			SKG_Debug( "SYNSL: not Players" )
			local rollList, listLog, lastMod, snap, snapLM, playerHash = unpack( message, 3, 8 )

			local myPlayerHash = SKG_HashRollList( 0, "Players" )

			-- First step... if our player hash is not the same as theirs, we
			-- should not attempt a sync. That needs to be true first.
			if ( myPlayerHash ~= playerHash ) then
				-- error
				SKG_Print( SKGL["SKG_ERROR_BADPLAYERHASH"] )
				return
			else
				-- next step: find out if both the passed list and ours
				-- have the same starting event (SYNSS)
				local oldestTID = SKG_GetOldestTID( listLog )
				local myOldestTID = SKG_GetOldestTID( db.Log[listName] )
				SKG_Debug( "his = " .. oldestTID )
				SKG_Debug( "mine = " .. myOldestTID )

				if ( oldestTID ~= myOldestTID ) then
					-- player needs to be able to "override" a bad starting
					-- point, and simply accept the other player's data as the
					-- correct data
					SKG.OverrideData = { 
						TID, listName, rollList, listLog, lastMod, snap, snapLM
					}
					StaticPopup_Show( "SKG_OVERRIDE_SYNSS" )
					return
				else
					-- Determine which list to send. If nlHash is 0, send
					-- an empty list. If it is not, we send the Snapshot
					-- list
					local tList = {}
					local lmList = {}
					if ( db.Log[listName][myOldestTID].nlHash ~= 0 ) then
						tList = SKG_tblCopy( db.Snapshot[listName] )
						lmList = SKG_tblCopy( db.SnapshotLM[listName] )
					end

					-- Create a combined list of transactions from both my
					-- and the sent list
					local transList = SKG_tblCopy( db.Log[listName] )

					for logTID, logTab in pairs( listLog ) do
						if ( transList[logTID] == nil ) then
							local tempT = SKG_tblCopy( logTab )
							transList[logTID] = tempT
						end
					end

					local result = SKG_RunTransactions( transList, tList, lmList )
					-- the results will be in tList
					if ( result == 0 ) then
						db.RollLists[listName] = tList
						db.Log[listName] = transList
						db.LastMod[listName] = lmList
					else
						SKG_Print( SKGL["SKG_ERROR_SYNCFAIL"] )
					end
					-- we still need to do LastMod stuff ???
				end
			end
			-- fix
			local lTID = SKG_GetLatestTID( db.Log[listName] )
			local newlHash = SKG_HashRollList( lTID, listName )
			SKG_AddLogEntry( TID, "SKG", "SYNSL", listName, newlHash )
		end
		SKG_Print( SKGL["SKG_MSG_SYNCCOMPLETE"] )
		SKG_ScrollFrame.selected = 0
		SKG_UpdateSPlayers()
		SKG_UpdateMainsChooser()
		SKG_UpdateToonChooser()
		SKG_UpdateCurrentList()
	end
end

SKG_OCR.SYNPL = function( sender, message )
	-- SYNPL
	-- Payload: (no payload)
	-- Description: This is the "Ping Lists" request, broadcast to the GUILD
	--		channel on login. The responses should come whispered as SYNAL (Ack Lists)
	--		messages.
	--
	-- Notes: This is to get around the fact that not everyone may have the lists
	--		that others do. This will give them at least the stub lists so they can
	--		then easily sync with others.

	-- Only respond if we're a syncer (rewrite for security)
	
	SKG_Debug( "Comm: SYNPL received" )

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- Don't bother if I'm the sender
		if ( sender ~= SKG.me ) then
			-- build result table
			local ResultSet = {}
			for k, v in pairs( SKG.db.profile.RollLists ) do
				tinsert( ResultSet, k )
			end

			-- Only send if we ourselves have any lists...
			if ( next( ResultSet ) ~= nil ) then
				SKG_SendComm( "SYNAL", ResultSet, "WHISPER", sender )
				SKG_Debug( "Comm: Sending SYNAL" )
			end
		end
	end
end

SKG_OCR.SYNAL = function( sender, message )
	-- SYNAL
	-- Payload: { List1, List2, ... }
	-- Description: The "Ack Lists" response to a SYNPL request. Whisper channel
	--		the response back with a list of all RollLists (exclude Players/Alts
	--		lists). 

	SKG_Debug( "Comm: SYNAL received" )
	local db = SKG.db.profile
	-- Don't bother logging these (jira SKG-145)
	-- local newTID = SKG_CreateTID( 0, 0, 0 )
	-- SKG_AddLogEntry( newTID, "SKG", "SYNAL", 0, 0 )

	for i = 1, #message do
		currList = message[i]
		if not ( db.RollLists[currList] ) then
			newTID = SKG_CreateTID( i, 0, 0 )

			SKG_Debug( "SYNAL:  New List created : " .. currList )

			db.RollLists[currList] = {}
			SKG_UpdateLists()

			-- If we're an admin, start off the log for our new list
			if ( SKG_SECURITY_LEVEL ~= "U" ) then
				db.LastMod[currList] = {}
				db.Log[currList] = {}
				db.Snapshot[currList] = {}
				db.SnapshotLM[currList] = {}
				SKG_AddLogEntry( newTID, currList, "SYNSS", 0, 0 )
			end
		end
	end
	-- force a quick update. This is more for someone who's security
	-- has just been upgraded than anything else...
	SKG_UpdateCurrentList()
end

------------------------------------------------
-- Security Commands 
------------------------------------------------

SKG_OCR.SECUR = function( sender, message )
	-- SECUR
	-- Payload: { Security Info Table }
	-- Description: Parse the sent Security Info, and see if my 
	--		permissions have changed. Alter my SV as appropriate.

	SKG_Debug( "Comm:  SECUR packet received" )

	local db = SKG.db.profile

	-- save the starting security level to compare with at the end
	local startLevel = SKG_SECURITY_LEVEL
	-- Whenever we get a new SECUR, we should once again assume we're just a user
	-- and let the packet prove us wrong
	if ( SKG_SECURITY_LEVEL ~= "A" ) then
		SKG_SECURITY_LEVEL = "U"
		local myRank = 20
		if ( message.RankBasedAuth == true ) then
			SKG_Debug( "SECUR: rba" )
			-- Rank based
			GuildRoster()
			for i = 1, GetNumGuildMembers( true ) do
				local gName, _, gRankIndex = GetGuildRosterInfo(i);
				if ( gName == SKG.me ) then
					myRank = gRankIndex
					break
				end
			end
			if ( myRank <= message.AuthRank ) then
				SKG_SECURITY_LEVEL = "S"
			end
		else
			SKG_Debug( "SECUR: man" )
			-- ManualAuth
			for i = 1, #message.ManualAuthList do
				if ( SKG.me == message.ManualAuthList[i] ) then
					SKG_SECURITY_LEVEL = "S"
					break
				end
			end
		end
		-- check if we're the proxy?
		if ( ( message.UseProxy == true ) and ( SKG.me == message.Proxy ) ) then
				SKG_Debug( "SECUR: pxy" )
				SKG_SECURITY_LEVEL = "P"
		end
	end

	local SI = SKG.SecurityInfo
	local dbSO = SKG.db.profile.Options.Security
	-- Get our Faction/Server/Guild info:
	local myFaction = UnitFactionGroup("player")
	local myServer = SKG_chopName( GetRealmName() )
	local myGuild = SKG_chopName( GetGuildInfo("player") )
	-- We should be saving the security info no matter what toon we're logged
	-- into at this point. Make sure all the parts of the tree exist
	if not ( dbSO.Policies ) then
		dbSO.Policies = {}
	end
	if not ( dbSO.Policies[myServer] ) then
		dbSO.Policies[myServer] = {}
	end
	if not ( dbSO.Policies[myServer][myFaction] ) then
		dbSO.Policies[myServer][myFaction] = {}
	end
	local dbSOP = dbSO.Policies[myServer][myFaction]

	-- New Security Model. Don't save literal Security information in
	-- the SV. Just local state variables.
	-- Check for extended security info. We were or probably a proxy of GL 
	-- on another toon.
	if ( ( SI.Proxy ) or 
		 ( SKG_SECURITY_LEVEL == "P" ) or
		 ( SKG_SECURITY_LEVEL == "A" ) ) then

		SI.Proxy = message.Proxy
		SI.RankBasedAuth = message.RankBasedAuth
		SI.AuthRank = message.AuthRank
		SI.UseProxy = message.UseProxy
		SI.ManualAuthList = {}
		SI.ManualAuthList = SKG_tblCopy( message.ManualAuthList )
		SKG:EnableModule( "SKGAdmin" )
		SKG_SF_InitValues()
	end

	SKG_Debug( "replacing security key" )
	dbSOP[myGuild] = SKG_encodeSECUR( message )
	
	if ( ( SKG_SECURITY_LEVEL == "U" ) or
		 ( SKG_SECURITY_LEVEL == "S" ) ) then
		SKG_FRAME_STATE = "Hidden"
		SKG_OptionsFrame:Hide()
 		SKG_HelpFrame:Hide()
		SKG_BidFrame:Hide()
		SKG_BidButtonFrame:Hide()
		SKG_ListFrame:Hide()
		SKG_SyncFrame:Hide()
		SKG_SecurityFrame:Hide()
	end

	local major, minor = SKG_getVersionComponents()
	SKG_TitleText:SetText( string.format( SKGL["SKG_LBL_TITLE_TEXT"] , 
		major, minor, SKG_SECURITY_LEVEL) )

	-- ok, so if we've changed security levels from "User" to anything else, all
	-- our SVs are going to be incomplete. Print a warning message to this effect,
	-- and reset us so we can resync cleanly. Oh, and send a SYNPL after so we at
	-- least get the list of current lists
	if ( ( startLevel == "U" ) and ( SKG_SECURITY_LEVEL ~= startLevel ) ) then
		SKG_Print( SKGL["SKG_MSG_RESETTINGSV"] )

		-- Clear out entries for old rolllists
		for rlName, rList in pairs( db.RollLists ) do
			db.RollLists[rlName] = nil
		end
		db.RollLists = nil
		db.LastMod = nil
		db.Log = nil
		db.Snapshot = nil
		db.SnapshotLM = nil
		db.RollLists = {}
		db.LastMod = {}
		db.Log = {}
		db.Log["SKG"] = {}
		db.Snapshot = {}
		db.SnapshotLM = {}

		SKG_SendComm( "SYNPL", {}, "GUILD", nil )
	end
end

------------------------------------------------
-- List Commands 
------------------------------------------------

SKG_OCR.LSTCR = function( sender, message )
	-- LSTCR
	-- Payload: { TID, ListName }
	-- Description: Creates a new list called "ListName"

	-- TODO : Redo security

	local db = SKG.db.profile
	local TID, listName = unpack( message, 1, 2 )

	if ( db.RollLists[listName] ) then
		-- We already have the list. Blah. Don't do anything.
	else
		SKG_Debug( "Comm:  New List created" )

		db.RollLists[listName] = {}
		SKG_UpdateLists()

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			SKG_AddLogEntry( TID, "SKG", "LSTCR", listName, 0 )

			-- And start off the log for our new list
			db.LastMod[listName] = {}
			db.Log[listName] = {}
			db.Snapshot[listName] = {}
			db.SnapshotLM[listName] = {}

			SKG_AddLogEntry( TID, listName, "SYNSS", 0, 0 )
		end
	end
end

SKG_OCR.LSTRM = function( sender, message )
	-- LSTRM
	-- Payload: { TID, ListName	}
	-- Description: Removes the list called "ListName"

	-- TODO: Redo security

	local db = SKG.db.profile
	local TID, listName = unpack( message, 1, 2 )

	if ( db.RollLists[listName] ) then
		SKG_Debug( "Comm:  Delete existing list" )

		db.RollLists[listName] = nil
		SKG_UpdateLists()
		SKG_ResetCurrentList()

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			db.Log[listName] = nil
			db.LastMod[listName] = nil
			db.Snapshot[listName] = nil
			db.SnapshotLM[listName] = nil

			SKG_AddLogEntry( TID, "SKG", "LSTRM", listName, 0 )
		end
	end
end

SKG_OCR.LSTRN = function( sender, message )
	-- LSTRN
	-- Payload: { TID, OldList, NewName }
	-- Description: Renames list "OldList" to "NewName"

	-- TODO : rewrite security

	local db = SKG.db.profile
	local TID, oldList, newName = unpack( message, 1, 3 )

	if ( db.RollLists[oldList] ) then
		SKG_Debug( "Comm:  Delete existing list" )

		db.RollLists[newName] = {}
		db.RollLists[newName] = db.RollLists[oldList]
		db.RollLists[oldList] = nil
		SKG_UpdateLists()
		SKG_ResetCurrentList()

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			-- Move the aux lists over as well
			db.Log[newName] = db.Log[oldList]
			db.LastMod[newName] = db.LastMod[oldList]
			db.Snapshot[newName] = db.Snapshot[oldList]
			db.SnapshotLM[newName] = db.SnapshotLM[oldList]

			db.Log[oldList] = nil
			db.LastMod[oldList] = nil
			db.Snapshot[oldList] = nil
			db.SnapshotLM[oldList] = nil

			local payload = { oldList, newName }
			SKG_AddLogEntry( TID, "SKG", "LSTRN", payload, 0 )
		end
	end
end

SKG_OCR.LSTBD = function( sender, message )
	-- LSTBD
	-- Payload: { listName, RollList }
	-- Description: This is what gets sent when a syncer broadcasts a list out
	--		to the users. As long as we're a user, we will pretty much
	--		unconditionally replace what we've got with what he's sent.

	-- At some point, might have to be wary of not being in a locked state
	if ( SKG_SECURITY_LEVEL == "U" ) then
		local listName, rList = unpack( message, 1, 2 )
		-- not going to bother even putting any checks and balances in here. Just do it.
		local db = SKG.db.profile
		if ( listName == "Players" ) then
			db.Players = nil
			db.Players = {}
			db.Players = SKG_tblCopy( rList )
		else
			db.RollLists[listName] = nil
			db.RollLists[listName] = {}
			db.RollLists[listName] = SKG_tblCopy( rList )
		end
		SKG_UpdateLists()
		SKG_UpdateSPlayers()
		SKG_UpdateToonChooser()
		SKG_UpdateCurrentList()
		SKG_Print( string.format( SKGL["SKG_MSG_LISTRECEIVED"], listName, sender ) )
	end
end

SKG_OCR.LSTCL = function( sender, message )
	-- LSTCL
	-- Payload: { TID, OldList, NewName }
	-- Description: Clones list "OldList" with "NewName"

	-- TODO : rewrite security

	local db = SKG.db.profile
	local TID, oldList, newName = unpack( message, 1, 3 )
	if ( db.RollLists[newName] ) then
		-- we've already got one
	else
		SKG_Debug( "Comm: Clone existing list" )

		db.RollLists[newName] = {}
		for key, val in pairs( db.RollLists[oldList] ) do
			db.RollLists[newName][key] = val
		end

		SKG_UpdateLists()
		SKG_ResetCurrentList()
		
		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			-- duplicate the "LastMod" table, but reset all values
			-- to the current timestamp
			db.LastMod[newName] = {}
			local dateStamp = strsub( TID, 1, 14 )
			for k, v in pairs( db.LastMod[oldList] ) do
				db.LastMod[newName][k] = dateStamp
			end

			-- if there was any log there leftover from something else, get
			-- initialize it
			db.Log[newName] = {}
			db.Snapshot[newName] = {}
			db.SnapshotLM[newName] = {}

			local payload = { oldList, newName }
			SKG_AddLogEntry( TID, "SKG", "LSTCL", payload, 0 )
			-- Init the new list's Log to a LSTCL event
			db.Log[newName] = {}
			local lHash, sHash = SKG_HashRollList( TID, newName )
			SKG_AddLogEntry( TID, newName, "LSTCL", 0, lHash )
		end
	end
end

------------------------------------------------
-- Player Commands 
------------------------------------------------

SKG_OCR.PLRAD = function( sender, message )
	-- PLRAD
	-- Payload: { TID, Player, Class, [PID] }
	-- Description: Adds "Player" to the list of valid "Players"
	--		If "PID" is present, "Player" is an alt of the player with id "PID"

	local PID = ""
	local TID, Player, Class = unpack( message, 1, 3 )
	if ( message[4] ) then
		PID = message[4]
	end

	SKG_Debug( "class is " .. Class )

	if ( ( PID == "" ) or ( PID == nil ) ) then
		-- We're inserting a main...
		local dblist = SKG.db.profile.Players

		-- create the player hash
		local PID = SKG_HashName( Player )
		local t = {
			name = Player,
			class = Class
		}
		dblist[PID] = t

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			local payload = { Player, Class }
			SKG_AddLogEntry( TID, "SKG", "PLRAD", payload, 0 )
		end
		SKG_UpdateMainsChooser()
	else	
		-- We're inserting an alt
		local dblist = SKG.db.profile.Alts

		if not dblist[PID] then dblist[PID] = {} end
		local t = {
			name = Player,
			class = Class
		}
		table.insert( dblist[PID], t )

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			local payload = { Player, Class, PID }
			SKG_AddLogEntry( TID, "SKG", "PLRAD", payload, 0 )
		end
	end
	SKG_UpdateSPlayers()
	SKG_UpdateToonChooser()
	SKG_UpdateCurrentList()
	SKG.isAlt = false
	SKG.selectedToon = ""
	SKG.mainPID = ""
	SKG.selectedClass = ""
	SKG.formSubmit = false

	if ( SKG.inRaid == true ) then
		SKG_UpdateRaiders()
	end
end

SKG_OCR.PLRIN = function( sender, message )
	-- PLRIN
	-- Payload: { TID, List, PID, [Position] }
	-- Description: Adds player with id "PID" into "List"
	-- Details: If "Position" is provided, the player will be added at that
	--		position, otherwise, they will be added to the bottom of the list.

	local Position = ""
	local db = SKG.db.profile
	local TID, listName, PID = unpack( message, 1, 3 )
	if ( message[4] ) then
		Position = message[4]
	end

	-- If we don't have this list, do nothing, or if our PID is invalid
	if ( db.RollLists[listName] ) then
		SKG_Debug( "pos = " .. Position )

		-- make sure our PID is valid too
		if ( ( PID == 0 ) or ( PID == nil ) or ( PID == "" ) ) then
			SKG_Print( SKGL["SKG_ERROR_BADPIDVALUE"] )
			return
		else
			local lastMod = strsub( TID, 1, 14 )
			local dblist = SKG.db.profile.RollLists[listName]

			tinsert( dblist, PID )
			local Payload = {}

			if ( Position == "" ) then
				Payload = { PID }
			else 
				Position = tonumber( Position )
				for i = #dblist, ( Position + 1 ), -1 do
					dblist[i] = dblist[i-1]
				end
				dblist[Position] = PID
				Payload = { PID, Position }
			end
			if ( SKG_SECURITY_LEVEL ~= "U" ) then
				db.LastMod[listName][PID] = lastMod
				local lHash, sHash = SKG_HashRollList( TID, listName )
				SKG_AddLogEntry( TID, listName, "PLRIN", Payload, lHash )
			end

			SKG_UpdateSPlayers()
			SKG_UpdateCurrentList()
			SKG.isAlt = false
			SKG.selectedToon = ""
			SKG.mainPID = ""
			SKG.selectedClass = ""
			SKG.formSubmit = false

			if ( SKG.inRaid == true ) then
				SKG_UpdateRaiders()
			end
		end
	end
end

SKG_OCR.PLRDL = function( sender, message )
	-- PLRDL
	-- Payload: { TID, PID, [Player] }
	-- Description:
	--		If no [Player] is provided, this takes the very broad stroke of
	--		removing the player represented by "PID" from all lists ("Players",
	--		"Alts" as well as any RollLists
	--
	--		If the optional [Player] is provided, this will remove the association
	--		from the "Alts" list for [Player] with "PID" (but not remove "PID" from
	--		any lists ("Players" or any RollLists).

	local Player = ""
	local TID, PID = unpack( message, 1, 2 )
	if ( message[3] ) then
		Player = message[3]
	end

	-- Check for valid PID
	if ( ( PID == 0 ) or ( PID == nil ) or ( PID == "" ) ) then
		SKG_Print( SKGL["SKG_ERROR_BADPIDVALUE"] )
		return
	else
		db = SKG.db.profile
		if ( Player == "" ) then
			-- Remove "PID" from all lists.
			-- alts...
			if ( db.Alts[PID] ~= nil ) then
				db.Alts[PID] = nil
			end
			-- loop through roll lists...
			for listName, listData in pairs( db.RollLists ) do
				local lIndex = 1
				while ( lIndex <= #db.RollLists[listName] ) do
					if ( db.RollLists[listName][lIndex] == PID ) then
						tremove( db.RollLists[listName], lIndex )
						if ( SKG_SECURITY_LEVEL ~= "U" ) then
							db.LastMod[listName][PID] = nil
							local lHash, sHash = SKG_HashRollList( TID, listName )
							SKG_AddLogEntry( TID, listName, "PLRDL", message, lHash )
						end
						break
					end
					lIndex = lIndex + 1
				end
			end
			-- and Players list
			db.Players[PID] = nil
			if ( SKG_SECURITY_LEVEL ~= "U" ) then
				SKG_AddLogEntry( TID, "SKG", "PLRDL", message, 0 )
			end
		else
			-- Only remove the alt [Player] from the "Alts" list
			local index = 1
			local notFound = true
			while ( ( index <= #db.Alts[PID] ) and ( notFound == true ) ) do
				if ( db.Alts[PID][index].name == Player ) then
					tremove( db.Alts[PID], index )
					notFound = false
				end
				index = index + 1
			end
			if ( next( db.Alts[PID] ) == nil ) then
				db.Alts[PID] = nil
			end
			if ( notFound == false ) then
				if ( SKG_SECURITY_LEVEL ~= "U" ) then
					SKG_AddLogEntry( TID, "SKG", "PLRDL", message, 0 )
				end
			end
		end
		SKG_ScrollFrame.selected = 0
		SKG_UpdateSPlayers()
		SKG_UpdateMainsChooser()
		SKG_UpdateToonChooser()
		SKG_UpdateCurrentList()
	end
end

SKG_OCR.PLRRM = function( sender, message )
	-- PLRRM	
	-- Payload: { TID, List, PID }
	-- Description:
	--		Removes player with id "PID" from "List"

	local TID, listName, PID = unpack( message, 1, 3 )
	local db = SKG.db.profile

	-- Check for valid PID
	if ( ( PID == 0 ) or ( PID == nil ) or ( PID == "" ) ) then
		SKG_Print( SKGL["SKG_ERROR_BADPIDVALUE"] )
		return
	else
		-- if we don't have that list, do nothing
		if ( db.RollLists[listName] ) then
			local index = 1
			local notFound = true
			while ( ( index <= #db.RollLists[listName] ) and ( notFound == true ) ) do
				if ( db.RollLists[listName][index] == PID ) then
					tremove( db.RollLists[listName], index )
					if ( SKG_SECURITY_LEVEL ~= "U" ) then
						db.LastMod[listName][PID] = nil
					end
					notFound = false
					break
				end
				index = index + 1
			end

			if ( notFound == false ) then
				-- this should *always* fire
				if ( SKG_SECURITY_LEVEL ~= "U" ) then
					local lHash, sHash = SKG_HashRollList( TID, listName )
					SKG_AddLogEntry( TID, listName, "PLRRM", PID, lHash )
				end
			end
	
			SKG_ScrollFrame.selected = 0
			SKG_UpdateSPlayers()
			SKG_UpdateMainsChooser()
			SKG_UpdateToonChooser()
			SKG_UpdateCurrentList()
		end
	end	
end

SKG_OCR.PLRMV = function( sender, message )
	-- PLRMV
	-- Payload: { TID, List, PID, Direction }
	-- Description: Moves player identified by PID in"Direction" in "List"
	-- Details: "Direction" will be either "UP" or "DOWN"

	local TID, listName, PID, Direction = unpack( message, 1, 4 )

	local root = SKG.db.profile

	-- Check for valid PID
	if ( ( PID == 0 ) or ( PID == nil ) or ( PID == "" ) ) then
		SKG_Print( SKGL["SKG_ERROR_BADPIDVALUE"] )
		return
	else
		-- if we don't have this list, do nothing
		if ( root.RollLists[listName] ) then
			local db = root.RollLists[listName]

			-- figure out "Position"
			local Position = 0
			for	i = 1, #db do
				if ( db[i] == PID ) then
					Position = i
					break
				end
			end

			if ( Position == 0 ) then
				-- This should never happen, but check anyway
			else
				-- If an admin clicks the move player button too quickly, it *IS* possible
				-- to get a player to move past the edges of a list. Add in checks for that
				-- here as well. jira SKG-151
				if ( ( Direction == "UP" ) and ( Position == 1 ) ) then
					SKG_Print( SKGL["SKG_ERROR_CANTMOVEHIGHER"] )
					return
				end
				if ( ( Direction == "DOWN" ) and ( Position == #db ) ) then
					SKG_Print( SKGL["SKG_ERROR_CANTMOVELOWER"] )
					return
				end

				if ( Direction == "UP") then
					db[Position], db[Position - 1] = db[Position - 1], db[Position]
				else
					db[Position], db[Position + 1] = db[Position + 1], db[Position]
				end

				-- We want to update if the displayed list is the one being modified...
				if ( listName == SKG_ActiveListTitle:GetText() ) then
					-- We're displaying the list being modified.

					-- If the currently selected player was one of the ones that was
					-- moved, the selection should 'follow' the move.
					local SKGs = tonumber( SKG_ScrollFrame.selected )
					if ( Direction == "UP" ) then
						if ( SKGs == Position ) then SKG_ScrollFrame.selected = Position - 1 end
						if ( SKGs == ( Position - 1 )) then SKG_ScrollFrame.selected = Position end
					else
						if ( SKGs == Position ) then SKG_ScrollFrame.selected = Position + 1 end
						if ( SKGs == ( Position + 1 )) then SKG_ScrollFrame.selected = Position end
					end
					SKG_Debug( "select = " .. SKG_ScrollFrame.selected )
					SKG_UpdateCurrentList()
				else
					-- We're displaying something else. Don't change a thing
				end

				if ( SKG_SECURITY_LEVEL ~= "U" ) then
					local payload = { PID, Direction }
					local lHash, sHash = SKG_HashRollList( TID, listName )
					SKG_AddLogEntry( TID, listName, "PLRMV", payload, lHash )
				end
			end
		end
	end
end

SKG_OCR.PLRSC = function( sender, message )
	-- PLRSC
	-- Payload : { TID,  listName, Player, { PID, PID, ... } }
	-- Description : Suicides "Player" in listName, using the list of PIDs
	--		as the list of people that are affected by the movement. This
	--		typically be anyone in the raid and/or reserved.
	--
	-- Details : "Player" should also be a PID, and will be the player to
	--		be suicided.

	local TID, listName, Player, PIDList = unpack( message, 1, 4 )

	SKG_Debug( "In PLRSC" )
	SKG_Debug( "TID is " .. TID )
	SKG_Debug( "list is " .. listName )
	SKG_Debug( "Player is " .. Player )
	SKG_Debug( "affectedlist is " .. #PIDList .. " entries " )
	local db = SKG.db.profile
	-- don't do anything unless we actually have this list
	if ( db.RollLists[listName] ) then
		-- First, lets create a list of the positions in "ListName" that are
		-- going to move, in the order they appear in the list.
		rollList = db.RollLists[listName]
		local moveList = {}
		local mlIndex = 1
		local foundFirst = false
		for i = 1, #rollList do
			local j = 1
			local notfound = true
			while ( j <= #PIDList ) do
				if ( rollList[i] == PIDList[j] ) then
					notfound = false
					break
				end
				j = j + 1
			end
			if ( notfound == false ) then
				if ( rollList[i] == Player ) then
					foundFirst = true
				end
				-- Only add the person to the moveList if we've already found
				-- the player to be suicided (including that player as well)
				if ( foundFirst == true ) then
					moveList[mlIndex] = i
					mlIndex = mlIndex + 1
				end
			end
		end

		-- At this point, moveList should contain, in order, the positions to move
		SKG_Debug( "our movelist contains " .. #moveList .. " entries." )

		-- Create the datestamp to use for all the LastMod fields.
		local dateStamp = strsub( TID, 1, 14 )
		local lmList = {}
		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			lmList = SKG.db.profile.LastMod[listName]
		end

		local moveLen = #moveList
		local SuicideTemp = rollList[moveList[1]]
		local moveIndex = 1
		while ( moveIndex < moveLen ) do
			SKG_Debug( "position in list = " .. moveList[moveIndex] )
			local movingPID = rollList[moveList[moveIndex + 1]]
			SKG_Debug( "movingPID = " .. movingPID )
			rollList[moveList[moveIndex]] = movingPID

			-- and lastmod stuff, if we're non-U
			if ( SKG_SECURITY_LEVEL ~= "U" ) then
				SKG_Debug( "old lastmod entry is " .. lmList[movingPID] )
				lmList[movingPID] = dateStamp
				SKG_Debug( "new lastmod entry is " .. lmList[movingPID] )
			end
			moveIndex = moveIndex + 1
		end
		rollList[moveList[moveIndex]] = SuicideTemp

		if ( SKG_SECURITY_LEVEL ~= "U" ) then
			lmList[SuicideTemp] = dateStamp
			local payload = { Player, PIDList }
			local lHash, sHash = SKG_HashRollList( TID, listName )
			SKG_AddLogEntry( TID, listName, "PLRSC", payload, lHash )
		end

		SKG_UpdateCurrentList()
	end
end

------------------------------------------------
-- Loot Commands 
------------------------------------------------

SKG_OCR.LOOTC = function( sender, message )
	-- LOOTC
	-- Payload: n/a
	-- Description: Clear Loot List

	SKG.LootItems = { }
	SKG_UpdateLootList()
end

SKG_OCR.LOOTI = function( sender, message )
	-- LOOTI
	-- Payload: { mobGUID, mobName, itemLink }
	-- Description: Add itemLink to the Loot List

	local mobGUID, mobName, itemLink = unpack( message, 1, 3 )
	-- For some reason, this got called with mobGUID being nil. I have no idea how that
	-- could possible have happeened. But... whatever, it did
	if ( mobGUID == nil ) then
		SKG_Debug( "omg. weird. mobGUID is nil. WTF?" )
	else
		SKG_Debug( "loot insert -> " .. itemLink .. " (" .. mobGUID .. ")" )
	end
	local t = {
		mGUID = mobGUID,
		mName = mobName,
		link = itemLink
	}

	tinsert( SKG.LootItems, 1, t )
	SKG_UpdateLootList()

	-- If we have autoopen set, pop open to the bid frame
	-- ONLY do this if we're NOT in combat.
	local inCombat = UnitAffectingCombat("player") -- returns 1 or nil
	if ( ( SKG.db.profile.Options.General.autoopen ) and ( inCombat == nil ) ) then
		SKG_OpenToBidFrame()
	end

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- If we have proper permissions open up the bidbuttonframe so
		-- a loot event can be handled.
		SKG_BBF_ClassFilterChooser:Enable()
		SKG_BBF_ArmorFilterChooser:Enable()
		SKG_BBF_OpenBidButton:Enable()
		SKG_BBF_CloseBidButton:Disable()
		SKG_BBF_ForceRetractButton:Disable()
		-- note: this does nothing (visible) if the parent frame is hidden
		SKG_BidButtonFrame:Show()
	else
		SKG_BidButtonFrame:Hide()
	end
end

SKG_OCR.LOOTR = function( sender, message )
	-- LOOTR
	-- Payload: { mobGUID, mobName, itemLink }
	-- Description: Remove itemLink that dropped off mobGUID from the Loot List

	SKG_Debug( "Received LOOTR" )
	local mobGUID, mobName, itemLink = unpack( message, 1, 3 )
	local found = 0
	for i = 1, #SKG.LootItems do
		if ( ( SKG.LootItems[i].link == itemLink ) and
			 ( SKG.LootItems[i].mName == mobName ) and
			 ( SKG.LootItems[i].mGUID == mobGUID ) ) then
			found = i
			break
		end
	end

	if ( found ~= 0 ) then
		tremove( SKG.LootItems, found )
	end

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		-- Disable. If there are still loot items, they will re-enable
		-- in UpdateLootList
		SKG_BBF_OpenBidButton:Disable()
		SKG_BBF_CloseBidButton:Disable()
		SKG_BBF_ArmorFilterChooser:Disable()
		SKG_BBF_ClassFilterChooser:Disable()
		SKG_BidButtonFrame:Hide()
	end
	SKG.selectedBidItem = 0
	SKG_UpdateLootList()
end

SKG_OCR.LOOTE = function( sender, message )
	-- LOOTE
	-- Payload: { TID, instance, mobName, itemName, itemID, itemRarity, pName, mPID }
	-- Description: Information regarding a completed loot event that finished with
	--		a winner (non-DE).

	if ( SKG_SECURITY_LEVEL ~= "U" ) then
		local TID, instanceName, mobName, itemName, itemID, itemRarity, playerName, mainPID = 
			unpack( message, 1, 8 )
		local payload = {
			instanceName, mobName,
			itemName, itemID, itemRarity,
			playerName, mainPID
		}		
		SKG_AddLogEntry( TID, "Loot", "LOOTE", payload, 0 )
	end
end

------------------------------------------------
-- Bidding Commands 
------------------------------------------------

SKG_OCR.BIDRT = function( sender, message )
	-- BIDRT
	-- Payload: bidder
	-- Description: Retract bidder from the current bid list

	local bidder = message

	local removed = 0
	if ( SKG.CurrBidList ~= "" ) then
		-- we've got a valid roll list
		-- loop through current bidders
		for k, v in pairs( SKG.Bidders ) do
			if ( v == bidder ) then
				removed = k
				break
			end
		end
		if ( removed ~= 0 ) then
			SKG.Bidders[removed] = nil
			SKG_UpdateBidderPanel()

			-- If I was the bidder, re-enable my bid button
			if ( SKG.me == bidder ) then
				SKG_BF_BidButton:Enable()
				SKG_BF_RetractButton:Disable()
			end

			-- If I'm the event handler, announce...
			if ( SKG.ManagingBidEvent == true ) then
				SendChatMessage( string.format( SKGL["SKG_MSG_RETRACTNOTIFY"] , bidder ), "RAID" )
				-- if we have an empty list, send a different message
				if ( next( SKG.Bidders ) == nil ) then
					SendChatMessage( SKGL["SKG_MSG_NOBIDDERS"], "RAID" )
				else
					-- find our high bidder
					local highBidPos = 0
					local highBidder = ""
					for	k, v in pairs( SKG.Bidders ) do
						if ( highBidPos == 0 ) then
							highBidPos = k
							highBidder = v
						else
							if ( k < highBidPos ) then
								highBidPos = k
								highBidder = v
							end
						end
					end
					-- notify
					SendChatMessage( string.format( SKGL["SKG_MSG_HIGHBIDDER"], highBidder ), "RAID" )
				end
			end
		else
			if ( SKG.ManagingBidEvent == true ) then
				-- our retractor wasn't in the list, let him know.
				SendChatMessage( SKGL["SKG_ERROR_CANTRETRACT"], "WHISPER", nil, bidder )
			end
		end
	end
end

SKG_OCR.BIDBD = function( sender, message )
	-- BIDBD
	-- Payload: bidder
	-- Description: Incoming bid on whatever the current item is, from "bidder"

	local bidder = message

	SKG_Debug( "Received bid event" )

	-- Handle a bunch of error checking up front
	if ( SKG.ManagingBidEvent ~= true ) then
		-- I'm not the loot event handler. 
		return
	end
	SKG_Debug( "I'm the event handler" )

	local foundInRaid = 0
	if ( #SKG.Raiders == 0 ) then
		-- I'm not in a raid
		return
	else
		-- we're in a raid, so find our man
		for i = 1, #SKG.Raiders do
			if ( SKG.Raiders[i].name == bidder ) then
				-- he's in our raid
				foundInRaid = i
				break
			end
		end
		if ( foundInRaid == 0 ) then
			SendChatMessage( SKGL["SKG_ERROR_NOTINMYRAID"], "WHISPER", nil, bidder )
			return
		end
	end
	SKG_Debug( "We're in a raid, and we found our bidder" )

	if ( SKG.CurrBidList == "" ) then
		-- we don't have a bid list. This really shouldn't even happen. Flag an
		-- error the admin, and return
		SKG_Print( SKGL["SKG_ERROR_NOBIDLIST"] )
		return
	end
	SKG_Debug( "current bid list is " .. SKG.CurrBidList )	

	local db = SKG.db.profile.RollLists[SKG.CurrBidList]
	if ( next( db ) == nil ) then
		-- somehow, our roll list no entries. *cry*
		SKG_Debug( "How does our roll list have no entries?" )
		return
	end

	local inList = 0
	for i = 1, #db do
		if ( db[i] == SKG.Raiders[foundInRaid].mPID ) then
			inList = i
			break
		end
	end
	if ( inList == 0 ) then
		SendChatMessage( SKGL["SKG_ERROR_NOTINBIDLIST"], "WHISPER", nil, bidder )
		SKG_Print( SKGL["SKG_ERROR_INVALIDBIDDER"] .. " " .. bidder )
		return
	end
	SKG_Debug( "We've got a valid bidder in our current roll list" )

	-- Looking good so far. Check our loot filter
	local foundClass = 0
	local bClass, enClass = UnitClass( bidder )
	if ( bClass ) then
		for i = 1, SKG.numClasses do
			if ( enClass == SKG_enClassNames[i] ) then
				foundClass = i
				break
			end
		end
	else
		-- Somehow, the dude didn't return a valid class. He's either not in our
		-- raid (and I can't find the class, or some other weirdness
		return
	end
	if ( foundClass == 0 ) then
		-- somehow, we couldn't find this guy's class in our classlist. Bad. Ignore
		return
	end
	SKG_Debug( "We've got a valid class back from the bidder" )

	SKG_Debug( SKG.BidFilter )
	SKG_Debug( "foundClass = " .. foundClass )
	if ( strsub( SKG.BidFilter, foundClass, foundClass ) ~= "1" ) then
		-- he did not pass our class filter. let him know and bail
		SendChatMessage( SKGL["SKG_ERROR_BIDFILTERED"], "WHISPER", nil, bidder )
		return
	end
	SKG_Debug( "He's made it through our loot filter" )

	-- Has this guy already bid?
	if ( SKG.Bidders[inList] ) then
		-- yikes, he's already bid. Don't let him bid again
		SendChatMessage( SKGL["SKG_ERROR_ALREADYBID"], "WHISPER", nil, bidder )
		return
	end
	SKG_Debug( "He hasn't bid yet" )

	-- If we've made it this far, we're good. THrow caution to the wind and
	-- process the bid

	local payload = { "ADDBIDDER", inList, bidder }
	SKG_SendComm( "SPECL", payload, "RAID", nil )
end

SKG_OCR.BIDCL = function( sender, message )
	-- BIDCL
	-- Payload: { TID, mobGUID, mobName, ItemLink }
	-- Description: Close the bidding for "ItemLink"

	local TID, mobGUID, mobName, itemLink = unpack( message, 1, 4 )

	-- If I'm the one handling a bid event, take care of the end processing
	if ( SKG.ManagingBidEvent == true ) then
		SKG_Debug( "  I'm the event handler" )
		-- do we have any bidders?
		if ( next( SKG.Bidders) ~= nil ) then
			-- we need to:
			-- find out high bidder
			local highBidPos = 0
			local highBidder = ""
			for k, v in pairs( SKG.Bidders ) do
				if ( highBidPos == 0 ) then
					highBidPos = k
					highBidder = v
				else
					if ( k < highBidPos ) then
						highBidPos = k
						highBidder = v
					end
				end
			end

			-- Find the winner's group # (jira SKG-146)
			local highBidderGroup = 0
			for i = 1, 40 do
				local currUnit = "raid" .. i
				if ( UnitExists( currUnit ) ) then
					local rName, _, rGroup = GetRaidRosterInfo( i ) 
					if ( rName ~= nil ) then
						if ( rName == highBidder ) then
							highBidderGroup = rGroup
						end
					end
				end
			end
						
			-- Announce the winner
			SendChatMessage( string.format( SKGL["SKG_MSG_WINNERANNOUNCE"], itemLink, highBidder, highBidderGroup, highBidPos ), "RAID" )

			-- send a suicide event

			-- Create TID
			local newTID = SKG_CreateTID( 0, 0, 0 )
			local listName = SKG.CurrBidList
			-- create our list of affected PIDs from SKG.Raiders()
			local affectedList = {}
			local winnerPID = 0
			for i = 1, #SKG.Raiders do
				tinsert( affectedList, SKG.Raiders[i].mPID )
				-- while we're in here, figure out which mPID belongs to the winner
				if ( SKG.Raiders[i].name == highBidder ) then
					winnerPID = SKG.Raiders[i].mPID
				end
			end
			-- Add in any reserved players
			for i = 1, #SKG.ReservedPlayers do
				tinsert( affectedList, SKG.ReservedPlayers[i] )
			end
			-- send off the suicide event
			local Payload = { newTID, listName, winnerPID, affectedList }
			SKG_SendComm( "PLRSC", Payload, "GUILD", nil )

			-- and .... the LOOTE event
			local eTID = SKG_CreateTID( 1, 0, 0 )
			-- where are we?
			local instanceName = GetRealZoneText()
			-- get rarity, name and ID out of the itemlink
			local itemID = string.match( itemLink, "item:(%d+)" )
			local itemName, _, itemRarity = GetItemInfo( itemLink ) 
		
			local Payload = {
				eTID,				-- timestamp of bid close event + index
				instanceName,		-- instance
				mobName,			-- mob
				itemName,			-- item's name
				itemID,				-- item id
				itemRarity,			-- quality
				highBidder,			-- toon that won
				winnerPID			-- toon's mainPID
			}
			SKG_SendComm( "LOOTE", Payload, "GUILD", nil )
		else
			-- Give the DE message
			SendChatMessage( SKGL["SKG_MSG_NOWINNER"], "RAID" )
		end
	end	

	-- for most people, this will just clear out the bidder list and remove
	-- the item from the biddable items list. Also, disable buttons as appropriate
	SKG.Bidders = {}
	SKG.CurrBidList = ""
	SKG_UpdateBidderPanel()

	-- Find the item in our list, and remove it
	for i = 1, #SKG.LootItems do
		if ( mobGUID == SKG.LootItems[i].mGUID ) then
			if ( SKG_SmartCompareItemlinks(
					SKG.LootItems[i].link, itemLink ) == true ) then
				tremove( SKG.LootItems, i )
				break
			end
		end
	end

	-- Reset the loot filter to "All"
	SKG.LootFilter = "1111111111"
	SKG.BidFilter = "1111111111"
	SKG_SetLootFilter ( "1111111111" )

	-- Disable. If there are still loot items, they will re-enable
	-- in UpdateLootList
	SKG_BBF_OpenBidButton:Disable()
	SKG_BBF_CloseBidButton:Disable()
	SKG_BBF_RemoveItemButton:Enable()
	SKG_BBF_ArmorFilterChooser:Disable()
	SKG_BBF_ClassFilterChooser:Disable()
	SKG_BidButtonFrame:Hide()
	SKG.selectedBidItem = 0

	SKG_UpdateLootList()

	SKG_BF_BidButton:Disable()
	SKG_BF_RetractButton:Disable()

	SKG.ManagingBidEvent = false
end

SKG_OCR.BIDOP = function( sender, message )
	-- BIDOP
	-- Payload: { TID, ItemLink, Filter, BidList }
	-- Description: Open the bidding for "ItemLink", filtered as appropriate by
	--		"Filter". For admin purposes, list being used (BidList) is also passed

	local TID, itemLink, filter, bidList = unpack( message, 1, 4 )
	SKG_Debug( "Received BIDOP event" )

	SKG.CurrBidList = bidList
	-- Reset any old bidders list, if there's something left over
	SKG.Bidders = nil
	SKG.Bidders = {}
	SKG_UpdateBidderPanel()

	-- If we pass the loot filter, enable our bid button
	local zClass, enClass = UnitClass( "player" )
	local found = 0
	for i = 1, SKG.numClasses do
		if ( enClass == SKG_enClassNames[i] ) then
			found = i
			break
		end
	end
	if ( found ~= 0 ) then
		if ( strsub( filter, found, found ) == "1" ) then
			-- we pass, enable the button
			SKG_BF_BidButton:Enable()
		else
			-- we don't pass, make sure its disabled
			SKG_BF_BidButton:Disable()
		end
	end

	-- If we have proper permissions (eventually), enable the appropriate buttons
	SKG_BBF_OpenBidButton:Disable()
	SKG_BBF_RemoveItemButton:Disable()
	SKG_BBF_CloseBidButton:Enable()
	SKG_BBF_ArmorFilterChooser:Disable()
	SKG_BBF_ClassFilterChooser:Disable()
end

------------------------------------------------
-- Special Commands
------------------------------------------------

SKG_OCR.SPECL = function( sender, message )

	-- This token is for very specific things that needed to be passed between
	-- addons that didn't fit in anywhere else, or were particular "end results"
	-- of some transactions already processed.
	--
	-- The payload passed to this function can be generalized as:
	--	{ "SPECIALTOKEN", ... }
	--
	-- The specific arguments that follow the special token will vary depending on
	-- the token. For now, I'm going to document them inline here...
	--
	-- { "ADDBIDDER", inList, bidder }
	--		inList		- position in the list to insert
	--		bidder		- the name to insert

	local SpecialToken = message[1]

	if ( SpecialToken == "ADDBIDDER" ) then
		local inList, bidder = unpack( message, 2, 3 )

		SKG.Bidders[inList] = bidder
		SKG_UpdateBidderPanel()

		-- If "bidder" is me, then hide my bid button, show the retract button
		if ( bidder == SKG.me ) then
			SKG_BF_BidButton:Disable()
			SKG_BF_RetractButton:Enable()
		end
			
		-- If we happen to the be loot handler, find our high bidder and announce
		if ( SKG.ManagingBidEvent == true ) then
			local highBidPos = 0
			local highBidder = ""
			for k, v in pairs( SKG.Bidders ) do
				if ( highBidPos == 0 ) then
					highBidPos = k
					highBidder = v
				else
					if ( k < highBidPos ) then
						highBidPos = k
						highBidder = v
					end
				end
			end
	
			SendChatMessage( string.format( SKGL["SKG_MSG_BIDNOTIFY"] , bidder, inList ), "RAID" )
			if ( highBidder == "" ) then
				SendChatMessage( string.format( SKGL["SKG_MSG_NOBIDDERS"], highBidder ), "RAID" )
			else 
				SendChatMessage( string.format( SKGL["SKG_MSG_HIGHBIDDER"], highBidder ), "RAID" )
			end
		end
	end
end

--[[
************************************************************************
CHANGELOG:

Date : 12/21/08
	Added Death Knight compatibility

Date : 10/26/08
	Moving RegisterComm to OnInit

Date : 10/25/08
	Initial attempt with Ace3 instead of LibRock. Added header/footer.
	Added functions to serialize/deserialize, and converted functions
	to a table SKG_OCR (OnCommReceive) which then has sub-functions for
	each token. Thats about as close to the old method as I can get.
************************************************************************
]]--